/*-
 * avail.c --
 *	Functions to check the status of the local machine to see
 *	if it can accept processes.
 *
 * Copyright (c) 1988 by the Regents of the University of California
 * Copyright (c) 1988 by Adam de Boor
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appears in all copies.  Neither the University of California nor
 * Adam de Boor makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 */
#ifndef lint
static char *rcsid =
"$Id: avail.c,v 1.17 88/11/17 20:42:18 adam Exp $ SPRITE (Berkeley)";
#endif lint

#include    "customsInt.h"

#undef NULL

#include    <sys/param.h>
#include    <sys/stat.h>
#include    <sys/conf.h>
#include    <sys/map.h>
#include    <nlist.h>
#include    <sys/file.h>
/*#include    <sys/time.h>*/
#include    <stdio.h>

#ifndef mapstart
/*
 * Apparently, Ultrix doesn't have these macros...
 */
#define mapstart(X)	(struct mapent *)((X)+1)
#define mapfree(X)	(X)->m_free
#define mapwant(X)	(X)->m_want
#define mapname(X)	((struct maplast *)(X))->m_nam
#endif /* mapstart */

static struct nlist kAddrs[] = {
{	"_nswapmap"	},   	/* Number of swap resource maps */
#define NSWAPMAP  	0
{	"_nswdev" 	},   	/* Number of swap devices   	*/
#define NSWDEV	  	1
{	"_swapmap"	},   	/* The swap resource maps   	*/
#define SWAPMAP	  	2
{	"_swdevt" 	},   	/* The swap devices 	    	*/
#define SWDEVT	  	3
{	"_avenrun"	},   	/* Load averages    	    	*/
#define AVENRUN	  	4
{	""  	  	}
};

static int  	  	kmem;	    	/* Descriptor to /dev/kmem */
static int  	  	kbd;	    	/* Descriptor to /dev/kbd */
#ifdef FSCALE
static int  	  	maxLoad = 0.5*LOADSCALE;
#else
static double	    	maxLoad = 0.5;
#endif /* FSCALE */
static int  	  	minSwapFree=25;	/* 75% used swap => not available */
#ifndef NO_IDLE
static int  	  	minIdle=15*60;	/* Keyboard must have been idle at
					 * least 15 minutes (when this is
					 * 15 * 60)... */
#else
static int  	    	minIdle = 0;	/* Don't consider idle time */
#endif /* NO_IDLE */
static int  	  	maxImports = 2;	/* Largest number of imported jobs */

#ifdef FSCALE
static long 	  	avenrun[3];	/* Load averages */
#else
static double		avenrun[3];	/* Load averages */
#endif /* FSCALE */

static int  	  	swblocks; 	/* Total swap space available */
static int  	  	nSwapMap; 	/* Number of entries in the swap map */
static off_t	  	swapMapAddr;	/* Address in sysspace of the map */
static struct map 	*swapMap; 	/* Space for swap map */
static int  	  	swapMapSize;	/* Size of the swap map (bytes) */

static Rpc_Event  	availEvent;	/* Event for checking availability */
static struct timeval	availInterval;	/* Interval at which checks should
					 * be made for the availability of
					 * this host. */


/*-
 *-----------------------------------------------------------------------
 * Avail_Send --
 *	Send the availability of the local host.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	An availability packet is sent to the master.
 *
 *-----------------------------------------------------------------------
 */
Boolean
Avail_Send ()
{
    Avail   	  avail;
    static int	  sending = 0;	    /* XXX: A kludge to prevent endless
				     * recursion. At times, for no reason I've
				     * been able to determine, the avail event
				     * will be triggered during the call to
				     * CUSTOMS_AVAIL (hard to believe since
				     * the timeout for the avail event is
				     * twice as long as the total for the
				     * rpc, but...). Once it starts, it
				     * continues and the calls never seem to
				     * complete. To prevent this, we use a
				     * static flag and don't send anything
				     * if a call is already being sent. */
    if (sending) {
	return(FALSE);
    } else {
	sending = 1;
    }
    
    avail.addr =  	localAddr.sin_addr;
    avail.interval = 	availInterval;
    avail.avail = 	Avail_Local(AVAIL_EVERYTHING);

    if (verbose) {
	printf ("Localhost %s available\n", avail.avail ? "not" : "is");
	fflush(stdout);
    }

    if (!Elect_InProgress() &&
	(Rpc_Call(udpSocket, &masterAddr, (Rpc_Proc)CUSTOMS_AVAIL,
		  sizeof(avail), (Rpc_Opaque)&avail,
		  0, (Rpc_Opaque)0,
		  CUSTOMSINT_NRETRY, &retryTimeOut) != RPC_SUCCESS)) {
		      Elect_GetMaster();
    }
    sending = 0;
    return (FALSE);
}

/*-
 *-----------------------------------------------------------------------
 * AvailSet --
 *	Set the availability criteria. Returns an OR of bits if the
 *	parameters are out-of-range.
 *
 * Results:
 *	Any of AVAIL_IDLE, AVAIL_SWAP, AVAIL_LOAD and AVAIL_IMPORTS
 *	or'ed together (or 0 if things are ok).
 *
 * Side Effects:
 *	The availabilty criteria are altered.
 *
 *-----------------------------------------------------------------------
 */
/*ARGSUSED*/
static void
AvailSet (from, msg, len, adPtr, fromRemote)
    struct sockaddr_in	*from;	    /* Address of sender */
    Rpc_Message	  	msg;	    /* Message for return */
    int	    	  	len;	    /* Length of criteria */
    Avail_Data	  	*adPtr;	    /* New criteria */
    Boolean 	  	fromRemote; /* TRUE if from remote call */
{
    int	    	  result;

    if (!Local(from)) {
	Rpc_Error(msg, RPC_ACCESS);
    } else if (len != sizeof(Avail_Data)) {
	Rpc_Error(msg, RPC_BADARGS);
    } else {
	result = 0;
	if ((adPtr->changeMask & AVAIL_IDLE)
#ifndef NO_IDLE
	    && (adPtr->idleTime > MAX_IDLE)
#endif /* NO_IDLE */
	    )
	{
		result |= AVAIL_IDLE;
	}
	if ((adPtr->changeMask & AVAIL_SWAP) &&
	    (adPtr->swapPct > MAX_SWAP))
	{
	    result |= AVAIL_SWAP;
	}
	if ((adPtr->changeMask & AVAIL_LOAD) &&
	    (adPtr->loadAvg < MIN_LOAD) &&
	    (adPtr->loadAvg != 0))
	{
	    result |= AVAIL_LOAD;
	}
	if ((adPtr->changeMask & AVAIL_IMPORTS) &&
	    (adPtr->imports < MIN_IMPORTS) &&
	    (adPtr->imports != 0))
	{
	    result |= AVAIL_IMPORTS;
	}
	if (result == 0) {
	    if (adPtr->changeMask & AVAIL_IDLE) {
		minIdle = adPtr->idleTime;
	    }
	    if (adPtr->changeMask & AVAIL_SWAP) {
		minSwapFree = adPtr->swapPct;
	    }
	    if (adPtr->changeMask & AVAIL_LOAD) {
#ifdef FSCALE
# if FSCALE == LOADSCALE
		maxLoad = adPtr->loadAvg;
# else
		maxLoad = ((double)adPtr->loadAvg/LOADSCALE) * FSCALE;
# endif /* FSCALE == LOADSCALE */
#else
		maxLoad = (double)adPtr->loadAvg/LOADSCALE;
#endif /* FSCALE */
	    }
	    if (adPtr->changeMask & AVAIL_IMPORTS) {
		maxImports = adPtr->imports;
	    }
	}
	adPtr->changeMask = result;
	adPtr->idleTime = minIdle;
	adPtr->swapPct = minSwapFree;
#ifdef FSCALE
# if FSCALE == LOADSCALE
	adPtr->loadAvg = maxLoad;
# else
  	adPtr->loadAvg = ((double)maxLoad/FSCALE) * LOADSCALE;
# endif
#else
    	adPtr->loadAvg = maxLoad * LOADSCALE;
#endif /* FSCALE */

	adPtr->imports = maxImports;
	if (fromRemote) {
	    Rpc_Return(msg, len, (Rpc_Opaque)adPtr);
	}
    }
}

/*-
 *-----------------------------------------------------------------------
 * AvailSetInterval --
 *	Alter the interval at which availability checks are made.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	The interval in availInterval is changed and availEvent is altered
 *	to reflect this change.
 *
 *-----------------------------------------------------------------------
 */
static void
AvailSetInterval (from, msg, len, intervalPtr)
    struct sockaddr_in	*from;
    Rpc_Message	  	msg;
    int	    	  	len;
    struct timeval	*intervalPtr;
{
    if (!Local(from)) {
	Rpc_Error(msg, RPC_ACCESS);
    } else if (len != sizeof(struct timeval)) {
	Rpc_Error(msg, RPC_BADARGS);
    } else if (intervalPtr->tv_sec < 5) {
	Rpc_Error(msg, RPC_BADARGS);
    } else {
	availInterval = *intervalPtr;
	Rpc_EventReset(availEvent, &availInterval);
	Rpc_Return(msg, 0, (Rpc_Opaque)0);
    }
}

/*-
 *-----------------------------------------------------------------------
 * Avail_Init --
 *	Initialize things for here...
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	We exit if can't initialize.
 *
 *-----------------------------------------------------------------------
 */
void
Avail_Init(criteria, checkTime)
    Avail_Data	  *criteria;	    /* Initial criteria */
    int	    	  checkTime;	    /* Initial check interval */
{
    struct swdevt *swapDevices,
		  *sw;
    int	    	  numSwapDevices;

    availInterval.tv_sec = checkTime ? checkTime : 10;
    availInterval.tv_usec = 0;

    if (nlist ("/vmunix", kAddrs) < 0) {
	printf ("/vmunix: could not read symbol table\n");
    } else if ((kmem = open ("/dev/kmem", O_RDONLY, 0)) < 0) {
	printf ("Could not open /dev/kmem\n");
#ifndef NO_IDLE
    } else if ((kbd = open ("/dev/kbd", O_RDONLY, 0)) < 0) {
	printf ("Could not open keyboard\n");
#endif /* NO_IDLE */
    } else {
	/*
	 * Find the total number of swap blocks available to the machine
	 * by summing the amounts in the swdevt descriptors
	 */
	lseek (kmem, (off_t)kAddrs[NSWDEV].n_value, L_SET);
	read (kmem, (char *)&numSwapDevices, sizeof (numSwapDevices));
	swapDevices = (struct swdevt *)malloc (numSwapDevices *
						  sizeof (struct swdevt));
	lseek (kmem, (off_t)kAddrs[SWDEVT].n_value, L_SET);
	read (kmem, (char *)swapDevices, numSwapDevices*sizeof(struct swdevt));
	for (swblocks = 0, sw = swapDevices;
	     numSwapDevices != 0;
	     sw++, numSwapDevices--) {
		 if (sw->sw_freed) {
		     swblocks += sw->sw_nblks;
		 }
	}
	free ((Address) swapDevices);

	/*
	 * Find and save the number and location of the swap maps for
	 * the local machine, then allocate enough room to hold them
	 * all, pointing 'swapMap' to the space.
	 */
	lseek (kmem, (off_t) kAddrs[NSWAPMAP].n_value, L_SET);
	read (kmem, (char *)&nSwapMap, sizeof (nSwapMap));
	lseek (kmem, (off_t) kAddrs[SWAPMAP].n_value, L_SET);
	read (kmem, (char *)&swapMapAddr, sizeof(swapMapAddr));

	printf ("%d swap blocks total allocated among %d maps at 0x%08x\n",
		  swblocks, nSwapMap, swapMapAddr);
	swapMapSize = nSwapMap * sizeof (struct map);

	swapMap = (struct map *) malloc (swapMapSize);

	availEvent = Rpc_EventCreate(&availInterval, Avail_Send,
				     (Rpc_Opaque)0);
	
	Rpc_ServerCreate(udpSocket, CUSTOMS_AVAILINTV, AvailSetInterval,
			 Swap_Timeval, Rpc_SwapNull, (Rpc_Opaque)0);
	Rpc_ServerCreate(udpSocket, CUSTOMS_SETAVAIL, AvailSet,
			 Swap_Avail, Swap_Avail, (Rpc_Opaque)TRUE);
	
	AvailSet(&localAddr, (Rpc_Message)0, sizeof(Avail_Data), criteria,
		 FALSE);
	return;
    }
    printf ("Can't initialize statistics gathering. Later days...\n");
    exit (1);
}

/*-
 *-----------------------------------------------------------------------
 * Avail_Local --
 *	Return TRUE if the local host is available for migration...
 *
 * Results:
 *
 * Side Effects:
 *
 *-----------------------------------------------------------------------
 */
int
Avail_Local(what)
    int	    what; 	/* Mask of things to check */
{
    struct stat	  	    kbStat;
    struct timeval	    now;
    register struct mapent  *mapEntry;
    struct mapent 	    *mapEnd;
    register int  	    free;

    /*
     * If an minimum idle time was specified, check to make sure the
     * keyboard idle time exceeds that.
     */
    if ((what & AVAIL_IDLE) && minIdle) {
#ifndef NO_IDLE
	fstat (kbd, &kbStat);
	gettimeofday (&now, (struct timezone *)0);
	
	if ((now.tv_sec - kbStat.st_atime) < minIdle) {
	    if (verbose) {
		printf ("Only %d seconds idle (%d min)\n",
			  now.tv_sec - kbStat.st_atime, minIdle);
		fflush(stdout);
	    }
	    return AVAIL_IDLE;
	}
#else
	;
#endif /* NO_IDLE */
    }

    /*
     * Either the machine has been idle long enough or the user didn't
     * specify an idle time, so now, if the user gave a free swap space
     * percentage beyond which the daemon may not go, tally up the total
     * free blocks in the swap map and see if it's too few.
     */
    if ((what & AVAIL_SWAP) && minSwapFree) {
	lseek (kmem, swapMapAddr, L_SET);
	read (kmem, (char *)swapMap, swapMapSize);
	
	mapEnd = (struct mapent *) &swapMap[nSwapMap];
	free = 0;
	for (mapEntry = mapstart(swapMap); mapEntry < mapEnd; mapEntry++) {
	    free += mapEntry->m_size;
	}
	if (((free * 100) / swblocks) < minSwapFree) {
	    if (verbose) {
		printf ("Only %d%% free (%d) swap blocks\n",
			  (free * 100) / swblocks, free);
		fflush(stdout);
	    }
	    return AVAIL_SWAP;
	}
    }

    /*
     * So far so good. Now if the user gave some maximum load average (note
     * that it can't be 0) which the daemon may not exceed, check all three
     * load averages to make sure that none exceeds the limit.
     */
    if ((what & AVAIL_LOAD) && maxLoad > 0) {
	lseek (kmem, (off_t) kAddrs[AVENRUN].n_value, L_SET);
	read (kmem, (char *)avenrun, sizeof (avenrun));
#ifdef ALL_LOAD
	if (avenrun[0] > maxLoad ||
	    avenrun[1] > maxLoad ||
	    avenrun[2] > maxLoad)
	{
		if (verbose) {
# ifdef FSCALE
		    printf ("loads: %f %f %f\n",
			    (double) avenrun[0]/FSCALE,
			    (double) avenrun[1]/FSCALE,
			    (double) avenrun[2]/FSCALE);
# else
		    printf ("loads: %f %f %f\n",
			    avenrun[0],
			    avenrun[1],
			    avenrun[2]);
# endif /* FSCALE */
		    fflush(stdout);
		}
		return AVAIL_LOAD;
	}
#else
	if (avenrun[0] > maxLoad) {
	    if (verbose) {
# ifdef FSCALE
		printf ("load: %f\n", (double) avenrun[0]/FSCALE);
# else
		printf ("load: %f\n", avenrun[0]);
# endif /* FSCALE */
	    }
	    return AVAIL_LOAD;
	}
#endif /* ALL_LOAD */
    }

    if ((what & AVAIL_IMPORTS) && maxImports &&
	(Import_NJobs() >= maxImports)) {
	    return AVAIL_IMPORTS;
    }

    /*
     * Great! This machine is available.
     */
    return 0;
}
